import functools
import time
import random
import logging
import importlib
import warnings
_exception = importlib.import_module(".exception", package="smawe_tools")


class Retrying(object):

    def __init__(
            self,
            func,
            stop_max_attempt_number=None,
            wait_random_min=None,
            wait_random_max=None,
            retry_on_exception=None,
    ):
        if not callable(func) and not isinstance(func, (classmethod, staticmethod)):
            raise ValueError("func param error")
        functools.update_wrapper(self, func)
        self._func = func
        self._f = None
        self._retry_on_exception = retry_on_exception if retry_on_exception is not None else Exception

        self._stop_max_attempt_number = stop_max_attempt_number if stop_max_attempt_number else 1

        self._wait_random_min = wait_random_min / 1000 if isinstance(wait_random_min, int) else 0
        self._wait_random_max = wait_random_max / 1000 if isinstance(wait_random_max, int) else 1
        if self._wait_random_max <= self._wait_random_min:
            raise ValueError("wait_random_min is greater than or equal to wait_random_max")

    def __get__(self, instance, owner=None):
        if isinstance(self._func, (staticmethod, classmethod)):
            self._f = self._func.__get__(instance, owner)
            functools.update_wrapper(self, self._f)
            return self

        self._f = functools.partial(self._func, instance)
        functools.update_wrapper(self, self._func)
        return self

    def __call__(self, *args, **kwargs):
        current_retry_num = 0
        while True:
            if current_retry_num > self._stop_max_attempt_number:
                raise _exception.MaxRetryError("Exceeded maximum retry count error")
            try:
                if current_retry_num:
                    logging.info("\033[1;34mThis is currently the {} retry\033[0m".format(current_retry_num))
                    time.sleep(random.uniform(self._wait_random_min, self._wait_random_max))
                if self._f:
                    return self._f(*args, **kwargs)
                return self._func(*args, **kwargs)
            except self._retry_on_exception:
                current_retry_num += 1


def retry(
    stop_max_attempt_number=None, wait_random_min=None, wait_random_max=None,
    retry_on_exception=None, **kwargs
):
    """
    异常重试装饰器, 0.3.6中添加了实例方法, 类方法, 静态方法的支持
    :param stop_max_attempt_number: 最大重试次数(默认为1次)
    :param wait_random_min: 重试间隔的随机等待最小时间(默认为0s), 单位毫秒
    :param wait_random_max: 重试间隔的随机等待最大时间(默认为1s), 单位毫秒
    :param retry_on_exception: 要重试的异常类型(默认为Exception)
    :return:
    """
    if retry_on_exception is None:
        retry_on_exception = kwargs.get('retry_exception')
        warnings.warn("retry_exception param was deprecated, please use retry_on_exception param",  category=DeprecationWarning, stacklevel=2)

    _kwargs = _merger_setting(
        stop_max_attempt_number=stop_max_attempt_number, wait_random_min=wait_random_min,
        wait_random_max=wait_random_max, retry_on_exception=retry_on_exception
    )
    return functools.partial(Retrying, **_kwargs)


def _merger_setting(**kwargs):
    new_dict = {}
    new_dict.update(kwargs)
    return new_dict


__all__ = ["retry"]



"""
字体颜色设置
    url: https://www.jianshu.com/p/a6791ba66e3d

Wechaty模块
    url: https://wechaty.readthedocs.io/zh_CN/latest/references/wechaty/

itchat
    url: https://itchat.readthedocs.io/zh/latest/tutorial/tutorial0/

windnd
    url: https://github.com/cilame/windnd
"""
from tkinter import *
from tkinter import ttk


def t():
    for i in range(100):
        pb.step()
        print(pb["value"])
        root.update()
        root.after(10)


root = Tk()
pb = ttk.Progressbar(root, orient="horizontal", mode="determinate", length=200, value=0, maximum=100.01)
pb.grid(column=1, row=0)
b = ttk.Button(root, text="start", command=t)
b.grid(column=0, row=0)

root.mainloop()



import socket
import re
import logging
import subprocess
import importlib
import sys
from types import ModuleType
from typing import Union
from smawe_tools.retrying import retry
_settings = importlib.import_module(".settings", "smawe_tools")


def _get_domain(url: str) -> str:
    try:
        _m = re.search(r"//(.*?)/", url.strip())
        return _m.group(1)
    except AttributeError:
        _m = re.search(r"//(.*)", url.strip())
        if _m:
            return _m.group(1)

    raise ValueError("please enter a valid url")


def get_ip(url: str = None, domain: str = None) -> _settings.List[str]:
    """
    get url or domain ip
    :param url: can be url or domain
    :param domain: website domain
    :return: list[str]
    """
    if url:
        url = url.strip()
        if "/" not in url:
            _domain = url
        else:
            _domain = _get_domain(url)
        try:
            (hostname, alias_list, ipaddr_list) = socket.gethostbyname_ex(_domain)
            return ipaddr_list
        except socket.gaierror as e:
            logging.error("please enter valid url")
            logging.error(e)

    if domain:
        domain = domain.strip()
        try:
            (hostname, alias_list, ipaddr_list) = socket.gethostbyname_ex(domain)
            return ipaddr_list
        except socket.gaierror as e:
            logging.error("please enter valid domain")
            logging.error(e)

    raise ValueError("please enter valid parameters")


@retry(stop_max_attempt_number=3, wait_random_min=1000, wait_random_max=2000)
def _install_requires(cmd: Union[str, _settings.List[str]]):
    subprocess.run(cmd)


def _run_cmd(cmd: Union[str, _settings.List[str]]) -> ModuleType:
    completed_p = subprocess.run(cmd)
    if not completed_p.returncode:
        logging.info("\033[1;34mThe dependent library is already installed\033[0m")
    else:
        _install_requires(cmd)
    try:
        import requests
    except ModuleNotFoundError as e:
        logging.error("Dependency library installation failed. Please check your network")
        logging.error("Failure Reason: {}".format(e))
        raise ValueError("The dependent library was not successfully installed: requests") from e
    else:
        return requests


def get_pubnet_ip() -> str:
    try:
        import requests
    except ModuleNotFoundError:
        logging.warning("\033[1;34mInstalling dependent libraries: requests\033[0m")
        if _settings.OS_NAME == "Windows":
            requests = _run_cmd("{} -m pip install requests".format(sys.executable))
        else:
            requests = _run_cmd("{} -m pip install requests".format(sys.executable).split())
    r = requests.get("https://www.httpbin.org/get")
    return r.json()['origin']


__all__ = ['get_ip', 'get_pubnet_ip']



import platform
import sys
import importlib
struct_ = importlib.import_module(".struct", package="smawe_tools")
try:
    from colorama import init as _init
    _init(autoreset=True)
except ModuleNotFoundError:
    pass

_IS_LESS_THEN_PY39 = float(".".join(platform.python_version_tuple()[:-1])) < 3.9
if _IS_LESS_THEN_PY39:
    import typing
    List = typing.List
else:
    List = list

# 'Linux', 'Darwin', 'Java', 'Windows'
OS_NAME = platform.system()


def modify_encoding(encoding="utf-8", language_code="en_US"):
    import _locale
    _locale._getdefaultlocale = lambda *args, **kwargs: (language_code, encoding)


sys.modules[__name__].__class__ = struct_.LoggingModule



def bubble_algorithm(array: list) -> list:
    """时间复杂度O(n^2), 对输入数组进行升序排序，返回排序后的数组副本"""
    array = array.copy()
    length = len(array)
    for i in range(length - 1):
        for j in range(length - 1 - i):
            if array[j] > array[j + 1]:
                array[j], array[j + 1] = array[j + 1], array[j]
    return array



import abc
import logging.handlers
import smtplib
from typing import List, Union, IO
from email.message import EmailMessage
import mimetypes
import re


class File(IO, metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, C):
        if cls is File:
            if any('read' in B.__dict__ for B in C.__mro__):
                return True
        return NotImplemented


SMTP_CONFIG = {
    'qq': {
        'host': "smtp.qq.com",
        'port': 465
    },
    '163': {
        'host': "smtp.163.com",
        'port': 465
    }
}


class ErrorLogger:
    """错误记录器，将日志记录发送到QQ邮箱."""

    def __new__(cls, *args, **kwargs):
        """
        返回一个name为error_logger的logging.Logger对象

        :param kwargs:
            '''

            from_addr: 发件人QQ邮箱, 未提供则为""

            to_addrs: 收件人QQ邮箱列表, 未提供则默认发送到from_addr

            subject: 邮件主题, 未提供则默认为"Program running"

            password: QQ授权码, 未提供则为""

            handler_level: 处理器记录级别, 未提供则默认为logging.INFO

            logger_level: 记录器记录级别, 未提供则默认为logging.INFO
            '''
        """
        from_addr = kwargs.pop("from_addr", "")
        to_addrs = kwargs.pop("to_addrs", [from_addr])
        subject = kwargs.pop("subject", "Program running")
        password = kwargs.pop("password", "")
        smtp_handler = logging.handlers.SMTPHandler(
            mailhost=("smtp.qq.com", 25), fromaddr=from_addr, toaddrs=to_addrs, subject=subject,
            credentials=(from_addr, password), secure=(), timeout=10
        )
        smtp_handler.setLevel(kwargs.pop("handler_level", logging.INFO))
        fmt = logging.Formatter("%(filename)s %(funcName)s %(levelname)s: (%(lineno)d)%(message)s")
        smtp_handler.setFormatter(fmt)
        error_logger = logging.getLogger("error_logger")
        error_logger.setLevel(kwargs.pop("logger_level", logging.INFO))
        error_logger.addHandler(smtp_handler)
        error_logger.email_helper = EmailHelper(user=from_addr, password=password)
        return error_logger


class EmailHelper:

    def __init__(self, user: str = None, password: str = None, host: str = None, port: int = None):
        """
        Support (QQ, 163) email
        :param user: username
        :param password: password
        :param host: smtp host
        :param port: smtp port
        """
        self.user = user
        brand = None
        pattern = '@([^\.]*)\.'
        match_obj = re.search(pattern, user)
        if match_obj is not None:
            brand = match_obj.group(1)

        config = None
        if brand is not None:
            _brand = brand.lower()
            try:
                config = SMTP_CONFIG[_brand]
            except KeyError:
                pass

        if config is not None:
            host = config.get('host')
            port = config.get('port')
            self.smtp_client = smtplib.SMTP_SSL(host=host, port=port)
            self.smtp_client.login(user=user, password=password)
            return

        if host is None:
            raise ValueError('Require following params: host')

        if port is None:
            port = 465

        self.smtp_client = smtplib.SMTP_SSL(host=host, port=port)
        self.smtp_client.login(user=user, password=password)

    def send_mail(
            self, message: str, to: List[str] = None, subject: str = None,
            file: Union[str, bytes, File] = None, file_name: str = None,
            files: List[Union[str, bytes, File]] = None, file_names: List[str] = None, strict=True
    ):
        """
        :param message: text message
        :param to: recipients, if is None, then is self
        :param subject: title
        :param file: open() -> file obj, used for single attachment
        :param file_name: 
        :param files: Used for multiple attachments
        :param file_names: List of file names corresponding to 'files'
        :param strict: default: True, when files and file_names inconsistent length, raised ValueError
        :return: 
        """
        msg = EmailMessage()
        msg['Subject'] = subject
        msg['From'] = self.user
        if to is None:
            to = [self.user]
        msg['To'] = ', '.join(to)
        msg.set_content(message)
        if file is not None:
            msg = self.set_attachment(msg, file, file_name)
        if files and file_names:
            if strict:
                if len(files) != len(file_names):
                    raise ValueError('files and file_names inconsistent length')

            for file, file_name in zip(files, file_names):
                msg = self.set_attachment(msg=msg, file=file, file_name=file_name)
        self.smtp_client.send_message(msg=msg)

    def set_attachment(self, msg: EmailMessage, file: Union[str, bytes, File], file_name: str) -> EmailMessage:
        if isinstance(file, str):
            attachment_content = open(file, 'rb').read()
            path = file
        elif isinstance(file, bytes):
            attachment_content = file
            path = ''
        else:
            attachment_content = file.read()
            if isinstance(attachment_content, str):
                attachment_content = attachment_content.encode()
            path = file.name
        ctype, encoding = mimetypes.guess_type(path)
        if ctype is None or encoding is not None:
            ctype = 'application/octet-stream'
        maintype, subtype = ctype.split('/', 1)
        msg.add_attachment(
            attachment_content,
            maintype=maintype,
            subtype=subtype,
            filename=file_name
        )
        return msg



from types import ModuleType
import logging
import sys


class LoggingModule(ModuleType):

    _name = "smawe_tools.settings"

    def __getattr__(self, name):
        if "__getattr__" in sys.modules["__main__"].__dict__:
            r = sys.modules["__main__"].__dict__["__getattr__"](name)
            if not r:
                raise AttributeError(f"module {type(self)._name!r} has no attribute {name!r}")
            return r
        raise AttributeError(f"module {type(self)._name!r} has no attribute {name!r}")

    def __getattribute__(self, name: str):
        return object.__getattribute__(self, name)

    def __setattr__(self, attr, value):
        if attr == "ENABLED_LOG":
            if value:
                logging.basicConfig(format="%(asctime)s:%(filename)s:%(threadName)s:%(levelname)s:%(message)s",
                                    level=logging.INFO)
        super().__setattr__(attr, value)
